import os
import tempfile
import typing as t
from pathlib import Path

import click
import gridfs
from superduper import CFG, logging
from superduper.backends.base.artifacts import ArtifactStore
from superduper.misc.colors import Colors
from tqdm import tqdm

from superduper_mongodb.utils import connection_callback


class MongoArtifactStore(ArtifactStore):
    """
    Artifact store for MongoDB.

    :param conn: MongoDB client connection
    :param name: Name of database to host filesystem
    :param flavour: Flavour of the artifact store
    """

    def __init__(
        self, conn, name: t.Optional[str] = None, flavour: t.Optional[str] = None
    ):
        super().__init__(name=name, conn=conn)
        if isinstance(conn, str):
            self.conn, url_name = connection_callback(conn, flavour)
            name = name or url_name
            self.name = f"_filesystem:{name}"
        else:
            self.conn = conn
            self.name = name
        self.filesystem = gridfs.GridFS(self.conn[self.name])

    def url(self):
        """Return the URL of the database."""
        return self.conn.HOST + ':' + str(self.conn.PORT) + '/' + self.name

    def drop(self, force: bool = False):
        """Drop the database.

        Please use with caution as this will delete all artifacts.

        :param force: If True, will not prompt for confirmation
        """
        if not force:
            if not click.confirm(
                f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
                f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
                'Are you sure you want to drop all artifacts? ',
                default=False,
            ):
                logging.warn('Aborting...')
        return self.conn.drop_database(self.name)

    def _exists(self, file_id):
        return self.filesystem.find_one({'file_id': file_id}) is not None

    def _delete_bytes(self, file_id: str):
        r = self.filesystem.find({'file_id': file_id})
        ids = [x._id for x in r]
        if not ids:
            raise FileNotFoundError(f'File not found in {file_id}')
        for _id in ids:
            self.filesystem.delete(_id)

    def get_bytes(self, file_id: str):
        """
        Get the bytes of the file from GridFS.

        :param file_id: The file_id of the file to get
        """
        cur = self.filesystem.find_one({'file_id': file_id})
        if cur is None:
            raise FileNotFoundError(f'File not found in {file_id}')
        return cur.read()

    def put_file(self, file_path: str, file_id: str):
        """Save file to GridFS.

        :param file_path: The path to the file to save
        :param file_id: The file_id of the file
        """
        path = Path(file_path)
        if path.is_dir():
            upload_folder(file_path, file_id, self.filesystem)
        else:
            _upload_file(file_path, file_id, self.filesystem)
        return file_id

    def get_file(self, file_id: str) -> str:
        """Download file from GridFS and return the path.

        The path is a temporary directory, `{tmp_prefix}/{file_id}/{filename or folder}`
        :param file_id: The file_id of the file to download
        """
        return _download(file_id, self.filesystem)

    def put_bytes(self, serialized: bytes, file_id: str):
        """
        Save bytes in GridFS.

        :param serialized: The bytes to save
        :param file_id: The file_id of the file
        """
        cur = self.filesystem.find_one({'file_id': file_id})
        if cur is not None:
            logging.warn(f"File {file_id} already exists")
            self._delete_bytes(file_id)
        return self.filesystem.put(serialized, filename=file_id, file_id=file_id)

    def disconnect(self):
        """Disconnect the client."""

        # TODO: implement me


def _upload_file(path, file_id, fs):
    """Upload file to GridFS.

    :param path: The path to the file to upload
    :param file_id: The file_id of the file
    :param fs: The GridFS object
    """
    logging.info(f"Uploading file {path} to GridFS with file_id {file_id}")
    path = Path(path)
    with open(path, 'rb') as file_to_upload:
        fs.put(
            file_to_upload,
            filename=path.name,
            file_id=file_id,
            metadata={"type": "file"},
        )


def upload_folder(path, file_id, fs, parent_path=""):
    """Upload folder to GridFS.

    :param path: The path to the folder to upload
    :param file_id: The file_id of the folder
    :param fs: The GridFS object
    :param parent_path: The parent path of the folder
    """
    path = Path(path)
    if not parent_path:
        logging.info(f"Uploading folder {path} to GridFS with file_id {file_id}")
        parent_path = os.path.basename(path)

    # if the folder is empty, create an empty file
    if not os.listdir(path):
        fs.put(
            b'',
            filename=os.path.join(parent_path, os.path.basename(path)),
            file_id=file_id,
            metadata={"is_empty_dir": True, 'type': 'dir'},
        )
    else:
        for item in os.listdir(path):
            item_path = os.path.join(path, item)
            if os.path.isdir(item_path):
                upload_folder(item_path, file_id, fs, os.path.join(parent_path, item))
            else:
                with open(item_path, 'rb') as file_to_upload:
                    fs.put(
                        file_to_upload,
                        filename=os.path.join(parent_path, item),
                        file_id=file_id,
                        metadata={"type": "dir"},
                    )


def _download(file_id, fs):
    """Download file or folder from GridFS and return the path.

    The path is a temporary directory, `{tmp_prefix}/{file_id}/{filename or folder}`

    :param file_id: The file_id of the file or folder to download
    :param fs: The GridFS object
    """
    download_folder = CFG.downloads.folder

    if not download_folder:
        download_folder = os.path.join(
            tempfile.gettempdir(), "superduper", "ArtifactStore"
        )

    save_folder = os.path.join(download_folder, file_id)
    os.makedirs(save_folder, exist_ok=True)

    file = fs.find_one({"file_id": file_id})
    if file is None:
        raise FileNotFoundError(f"File not found in {file_id}")

    type_ = file.metadata.get("type")
    if type_ not in {"file", "dir"}:
        raise ValueError(
            f"Unknown type '{type_}' for file_id {file_id}, expected file or dir"
        )

    if type_ == 'file':
        save_path = os.path.join(save_folder, os.path.split(file.filename)[-1])
        logging.info(f"Downloading file_id {file_id} to {save_path}")
        with open(save_path, 'wb') as f:
            f.write(file.read())
        logging.info(f"Downloaded file_id {file_id} to {save_path}")
        return save_path

    logging.info(f"Downloading folder with file_id {file_id} to {save_folder}")
    for grid_out in tqdm(fs.find({"file_id": file_id, "metadata.type": "dir"})):
        file_path = os.path.join(save_folder, grid_out.filename)
        if grid_out.metadata.get("is_empty_dir", False):
            if not os.path.exists(file_path):
                os.makedirs(file_path)
        else:
            directory = os.path.dirname(file_path)
            if not os.path.exists(directory):
                os.makedirs(directory)
            with open(file_path, 'wb') as file_to_write:
                file_to_write.write(grid_out.read())

    folders = os.listdir(save_folder)
    assert len(folders) == 1, f"Expected only one folder, got {folders}"
    folder_path = os.path.join(save_folder, folders[0])
    logging.info(f"Downloaded folder with file_id {file_id} to {folder_path}")
    return folder_path



import typing as t
from functools import cached_property

from superduper import CFG, logging
from superduper.backends.base.vector_search import BaseVectorSearcher, VectorItem

if t.TYPE_CHECKING:
    from superduper.components.vector_index import VectorIndex


class MongoAtlasVectorSearcher(BaseVectorSearcher):
    """Vector searcher implementation of atlas vector search.

    :param identifier: Unique string identifier of index
    :param collection: Collection name
    :param dimensions: Dimension of the vector embeddings
    :param measure: measure to assess similarity
    :param output_path: Path to the output
    """

    native_service: t.ClassVar[bool] = False

    def __init__(
        self,
        identifier: str,
        collection: str,
        dimensions: int,
        measure: t.Optional[str] = None,
        output_path: t.Optional[str] = None,
    ):
        import pymongo

        self.identifier = identifier
        vector_search_uri = CFG.cluster.vector_search.uri or CFG.data_backend
        assert vector_search_uri, "Vector search URI is required"
        db_name = vector_search_uri.split("/")[-1]
        self.database = getattr(pymongo.MongoClient(vector_search_uri), db_name)
        assert output_path
        self.output_path = output_path
        self.collection = collection
        self.measure = measure
        self.dimensions = dimensions
        self._is_exists = False

        self._check_if_exists(create=True)
        super().__init__(identifier=identifier, dimensions=dimensions, measure=measure)

    def __len__(self):
        pass

    @cached_property
    def index(self):
        """Return the index collection."""
        return self.database[self.collection]

    def is_initialized(self, identifier):
        """Check if vector index initialized."""
        return self._check_if_exists(create=False)

    @classmethod
    def from_component(cls, vi: "VectorIndex"):
        """Create a vector searcher from a vector index.

        :param vi: VectorIndex instance
        """
        from superduper.components.listener import Listener

        assert isinstance(vi.indexing_listener, Listener)
        assert vi.indexing_listener.select is not None
        path = collection = vi.indexing_listener.outputs

        return MongoAtlasVectorSearcher(
            identifier=vi.identifier,
            dimensions=vi.dimensions,
            measure=vi.measure,
            output_path=path,
            collection=collection,
        )

    def add(self, items: t.Sequence[VectorItem], cache: bool = False) -> None:
        """
        Add items to the index.

        :param items: t.Sequence of VectorItems
        """
        self._check_if_exists(create=True)

    def delete(self, ids: t.Sequence[str]) -> None:
        """Remove items from the index.

        :param ids: t.Sequence of ids of vectors.
        """

    def find_nearest_from_id(self, id: str, n=100, within_ids=None):
        """Find the nearest vectors to the given ID.

        :param id: ID of the vector
        :param n: number of nearest vectors to return
        :param within_ids: list of IDs to search within
        """
        h = self.index.find_one({"_id": id})[self.output_path]
        return self.find_nearest_from_array(h, n=n, within_ids=within_ids)

    def find_nearest_from_array(self, h, n=100, within_ids=None):
        """Find the nearest vectors to the given vector.

        :param h: vector
        :param n: number of nearest vectors to return
        :param within_ids: list of IDs to search within
        """
        self._check_if_exists(create=True)
        self._check_queryable()
        vector_search = {
            "index": self.identifier,
            "path": self.output_path,
            "queryVector": h,
            "numCandidates": n,
            "limit": n,
        }
        if within_ids:
            vector_search["filter"] = {"_id": {"$in": within_ids}}

        project = {
            "_id": 1,
            "_source": 1,
            "score": {"$meta": "vectorSearchScore"},
        }

        pipeline = [
            {"$vectorSearch": vector_search},
            {"$project": project},
        ]

        cursor = self.index.aggregate(pipeline)
        scores = []
        ids = []
        for vector in cursor:
            scores.append(vector["score"])
            ids.append(str(vector["_source"]))
        return ids, scores

    def _create_index(self):
        """Create a vector index in the data backend if an Atlas deployment."""
        if self.collection not in self.database.list_collection_names():
            logging.warn(
                f"Collection {self.collection} does not exist. " "Cannot create index."
            )
            return
        from pymongo.operations import SearchIndexModel

        definition = {
            "fields": [
                {
                    "type": "vector",
                    "numDimensions": self.dimensions,
                    "path": self.output_path,
                    "similarity": self.measure,
                },
            ]
        }

        search_index_model = SearchIndexModel(
            definition=definition,
            name=self.identifier,
            type="vectorSearch",
        )
        logging.info(
            f"Creating search index [{self.identifier}] on {self.collection} "
            f"-- Definition: {definition}"
        )
        result = self.index.create_search_index(model=search_index_model)
        return result

    def _check_if_exists(self, create=True):
        if self._is_exists:
            return True
        index = self._get_index()
        if bool(index):
            self._is_exists = True
        elif create:
            self._create_index()
        return self._is_exists

    def _check_queryable(self):
        index = self._get_index()
        if not index:
            raise FileNotFoundError(
                f"Index {self.identifier} does not exist in the collection "
                f"{self.collection}. Cannot perform query."
            )

        if not index.get("queryable"):
            raise FileNotFoundError(
                f"Index {self.identifier} is pending and not yet queryable. "
                "Please wait until the index is fully ready for queries."
                f"Cannot perform query. "
                f"You need to wait for the index to be queryable. "
                f"Index: {index}"
            )
        return True

    def _get_index(self):
        try:
            indexes = self.index.list_search_indexes()
        except Exception:
            return False

        indexes = [i for i in indexes if i["name"] == self.identifier]

        if not indexes:
            return None
        else:
            return indexes[0]



from .artifacts import MongoArtifactStore as ArtifactStore
from .data_backend import MongoDBDataBackend as DataBackend
from .metadata import MongoMetaDataStore as MetaDataStore
from .query import MongoQuery
from .vector_search import MongoAtlasVectorSearcher as VectorSearcher

__version__ = "0.5.1"

__all__ = [
    "ArtifactStore",
    "MongoQuery",
    "DataBackend",
    "MetaDataStore",
    "VectorSearcher",
]



import copy
import dataclasses as dc
import functools
import re
import typing as t
from collections import defaultdict

import pymongo
from bson import ObjectId
from superduper import CFG, logging
from superduper.backends.base.query import (
    Query,
    applies_to,
    parse_query as _parse_query,
)
from superduper.base.cursor import SuperDuperCursor
from superduper.base.document import Document, QueryUpdateDocument
from superduper.base.leaf import Leaf

if t.TYPE_CHECKING:
    from superduper.base.datalayer import Datalayer

_SPECIAL_CHRS: list = ['$', '.']

OPS_MAP = {
    '__eq__': '$eq',
    '__ne__': '$ne',
    '__lt__': '$lt',
    '__le__': '$lte',
    '__gt__': '$gt',
    '__ge__': '$gte',
    'isin': '$in',
}


def _serialize_special_character(d, to='encode'):
    def extract_character(s):
        pattern = r'<(.)>'
        match = re.search(pattern, s)
        if match:
            return match.group(1)
        return None

    if not isinstance(d, dict):
        return d

    new_dict = {}
    for key, value in d.items():
        new_key = key
        if isinstance(key, str):
            if to == 'encode':
                if key[0] in _SPECIAL_CHRS:
                    new_key = f'<{key[0]}>' + key[1:]
            elif to == 'decode':
                k = extract_character(key[:3])
                if k in _SPECIAL_CHRS:
                    new_key = k + key[3:]

        if isinstance(value, dict):
            new_dict[new_key] = _serialize_special_character(value, to=to)
        elif isinstance(value, list):
            new_dict[new_key] = [
                (
                    _serialize_special_character(item, to=to)
                    if isinstance(item, dict)
                    else item
                )
                for item in value
            ]
        else:
            new_dict[new_key] = value

    return new_dict


def parse_query(
    query, documents: t.Sequence[t.Dict] = (), db: t.Optional['Datalayer'] = None
):
    """Parse a string query into a query object.

    :param query: The query to parse.
    :param documents: The documents to query.
    :param db: The datalayer to use to execute the query.
    """
    _decode = functools.partial(_serialize_special_character, to='decode')
    documents = list(map(_decode, documents))
    return _parse_query(
        query=query,
        builder_cls=MongoQuery,
        documents=list(documents),
        db=db,
    )


class MongoQuery(Query):
    """A query class for MongoDB.

    This class is used to build and execute queries on a MongoDB database.
    """

    flavours: t.ClassVar[t.Dict[str, str]] = {
        'pre_like': r'^.*\.like\(.*\)\.(find|select)',
        'post_like': r'^.*\.(find|select)\(.*\)\.like(.*)$',
        'bulk_write': r'^.*\.bulk_write\(.*\)$',
        'outputs': r'^.*\.outputs\(.*\)',
        'missing_outputs': r'^.*\.missing_outputs\(.*\)$',
        'find_one': r'^.*\.find_one\(.*\)',
        'find': r'^.*\.find\(.*\)',
        'select': r'^.*\.select\(.*\)$',
        'insert_one': r'^.*\.insert_one\(.*\)$',
        'insert_many': r'^.*\.insert_many\(.*\)$',
        'insert': r'^.*\.insert\(.*\)$',
        'replace_one': r'^.*\.replace_one\(.*\)$',
        'update_many': r'^.*\.update_many\(.*\)$',
        'update_one': r'^.*\.update_one\(.*\)$',
        'delete_many': r'^.*\.delete_many\(.*\)$',
        'delete_one': r'^.*\.delete_one\(.*\)$',
        'other': '.*',
    }

    # Use to control the behavior in the class construction method within LeafMeta
    _dataclass_params: t.ClassVar[t.Dict[str, t.Any]] = {
        'eq': False,
        'order': False,
    }

    def _create_table_if_not_exists(self):
        return

    def dict(self, metadata: bool = True, defaults: bool = True, uuid: bool = True):
        """Return the query as a dictionary."""
        r = super().dict()
        r['documents'] = list(map(_serialize_special_character, r['documents']))
        return r

    @property
    def type(self):
        """Return the type of the query."""
        return defaultdict(
            lambda: 'select',
            {
                'update_many': 'update',
                'update_one': 'update',
                'delete_many': 'delete',
                'delete_one': 'delete',
                'bulk_write': 'write',
                'insert_many': 'insert',
                'insert_one': 'insert',
                'insert': 'insert',
                'outputs': 'select',
            },
        )[self.flavour]

    @property
    def _is_select_find(self):
        return self.parts and any([p[0] == 'select' for p in self.parts])

    def _prepare_inputs(self, inputs):
        if isinstance(inputs, BulkOp):
            return inputs.op
        if isinstance(inputs, (list, tuple)):
            return [self._prepare_inputs(i) for i in inputs]
        if isinstance(inputs, dict):
            return {k: self._prepare_inputs(v) for k, v in inputs.items()}
        return inputs

    def _execute_delete_one(self, parent):
        r = next(self.select_ids.limit(1)._execute(parent))
        self.table_or_collection.delete_one({'_id': r['_id']})._execute(parent)
        return [str(r['_id'])]

    def _execute_delete_many(self, parent):
        id_cursor = self.select_ids._execute(parent)
        ids = [r['_id'] for r in id_cursor]
        if not ids:
            return {}
        self.table_or_collection.delete_many({'_id': {'$in': ids}})._execute(parent)
        return [str(id) for id in ids]

    def _execute(self, parent, method='encode'):
        c = super()._execute(parent, method=method)
        import mongomock
        import pymongo

        if isinstance(c, (pymongo.cursor.Cursor, mongomock.collection.Cursor)):
            return SuperDuperCursor(
                raw_cursor=c,
                db=self.db,
                id_field='_id',
            )
        return c

    def _execute_bulk_write(self, parent):
        """Execute the query.

        :param db: The datalayer instance
        """
        assert self.parts[0][0] == 'bulk_write'
        operations = self.parts[0][1][0]
        for query in operations:
            assert isinstance(query, (BulkOp))

            query.is_output_query = self.is_output_query
            if not query.kwargs.get('arg_ids', None):
                raise ValueError(
                    'Please provided update/delete id in args',
                    r'all ids selection e.g `\{\}` is not supported',
                )

        collection = self.db.databackend.get_table_or_collection(self.table)
        bulk_operations = []
        bulk_update_ids = []
        bulk_delete_ids = []
        bulk_result = {'delete': [], 'update': []}
        for query in operations:
            operation = query.op

            bulk_operations.append(operation)
            ids = query.kwargs['arg_ids']
            if query.identifier == 'DeleteOne':
                bulk_result['delete'].append({'query': query, 'ids': ids})
                bulk_delete_ids += ids
            else:
                bulk_update_ids += ids
                bulk_result['update'].append({'query': query, 'ids': ids})

        result = collection.bulk_write(bulk_operations)
        if result.deleted_count != bulk_delete_ids:
            logging.warn(
                'Some delete ids are not executed',
                ', hence halting execution',
                'Please note the partially executed operations',
                'wont trigger any `model/listeners` unless CDC is active.',
            )
        elif (result.modified_count + result.upserted_count) != bulk_update_ids:
            logging.warn(
                'Some update ids are not executed',
                ', hence halting execution',
                'Please note the partially executed operations',
                'wont trigger any `model/listeners` unless CDC is active.',
            )
        return bulk_result, bulk_update_ids, bulk_delete_ids

    def filter(self, *args, **kwargs):
        """Return a query that filters the documents.

        :param args: The arguments to filter by.
        :param kwargs: Additional keyword arguments.
        """
        filters = {}
        for arg in args:
            if isinstance(arg, dict):
                filters.update(arg)
                continue
            if not isinstance(arg, Query):
                raise ValueError(f'Filter arguments must be queries, but got: {arg}')
            assert len(arg.parts) > 1, f'Invalid filter query: {arg}'
            key, (op, op_args, _) = arg.parts[-2:]

            if op not in OPS_MAP:
                raise ValueError(
                    f'Operation {op} not supported, '
                    f'supported operations are: {OPS_MAP.keys()}'
                )

            if not op_args:
                raise ValueError(f'No arguments found for operation {op}')

            value = op_args[0]

            filters[key] = {OPS_MAP[op]: value}

        query = self if self.parts else self.select()

        return type(self)(
            db=self.db,
            table=self.table,
            parts=[*query.parts, ('filter', (filters,), kwargs)],
        )

    def _execute_find(self, parent):
        return self._execute_select(parent)

    def _execute_replace_one(self, parent):
        documents = self.parts[0][1][0]
        trailing_args = list(self.parts[0][1][1:])
        kwargs = self.parts[0][2]

        schema = kwargs.pop('schema', None)

        replacement = trailing_args[0]
        if isinstance(replacement, Document):
            replacement = replacement.encode(schema)
        trailing_args[0] = replacement

        q = self.table_or_collection.replace_one(documents, *trailing_args, **kwargs)
        q._execute(parent)

    def _execute_find_one(self, parent):
        r = self._execute_select(parent)
        if r is None:
            return
        return Document.decode(r, db=self.db, schema=self._get_schema())

    def _execute_insert_one(self, parent):
        insert_part = self.parts[0]
        parts = self.parts[1:]
        insert_part = ('insert_many', [insert_part[1]], insert_part[2])

        self.parts = [insert_part] + parts
        return self._execute_insert_many(parent)

    def _execute_insert_many(self, parent):
        trailing_args = self.parts[0][1][1:]
        kwargs = self.parts[0][2]
        documents = self._prepare_documents()
        q = self.table_or_collection.insert_many(documents, *trailing_args, **kwargs)
        result = q._execute(parent)
        return result.inserted_ids

    def _execute_insert(self, parent):
        """Provide a unified insertion interface."""
        return self._execute_insert_many(parent)

    def _execute_update_many(self, parent):
        ids = [r['_id'] for r in self.select_ids._execute(parent)]
        filter = self.parts[0][1][0]
        trailing_args = list(self.parts[0][1][1:])
        update = {}
        kwargs = self.parts[0][2]

        # Encode update document
        for ix, arg in enumerate(trailing_args):
            if '$set' in arg:
                if isinstance(arg, Document):
                    update = QueryUpdateDocument.from_document(arg)
                else:
                    update = arg
                del trailing_args[ix]
                break

        filter['_id'] = {'$in': ids}

        try:
            table = self.db.load('table', self.table)
            schema = table.schema
        except FileNotFoundError:
            schema = None

        trailing_args.insert(
            0, update.encode(schema=schema) if isinstance(update, Document) else update
        )

        parent.update_many(filter, *trailing_args, **kwargs)
        return ids

    def drop_outputs(self, predict_id: str):
        """Return a query that removes output corresponding to the predict id.

        :param predict_ids: The ids of the predictions to select.
        """
        return self.db.databackend.drop_table_or_collection(
            f'{CFG.output_prefix}{predict_id}'
        )

    @applies_to('find', 'find_one', 'select', 'outputs')
    def add_fold(self, fold: str):
        """Return a query that adds a fold to the query.

        :param fold: The fold to add.
        """
        return self.filter(self['_fold'] == fold)

    @property
    @applies_to('insert_many', 'insert_one', 'insert')
    def documents(self):
        """Return the documents."""
        return super().documents

    @property
    def primary_id(self):
        """Return the primary id of the documents."""
        return '_id'

    def select_using_ids(self, ids: t.Sequence[str]):
        """Return a query that selects using the given ids.

        :param ids: The ids to select.
        """
        ids = [ObjectId(id) for id in ids]
        q = self.filter(self['_id'].isin(ids))
        return q

    @property
    def select_ids(self):
        """Select the ids of the documents."""
        if self._is_select_find:
            part = self.parts[0]
            new_part = ('select', ('_id',), part[2])
            return type(self)(
                db=self.db,
                table=self.table,
                parts=[new_part, *self.parts[1:]],
            )

        filter_ = {}
        if self.parts and self.parts[0][1]:
            filter_ = self.parts[0][1][0]
            if isinstance(filter_, str):
                filter_ = {}
        projection = {'_id': 1}
        coll = type(self)(table=self.table, db=self.db)
        return coll.find(filter_, projection)

    def select_ids_of_missing_outputs(self, predict_id: str):
        """Select the ids of documents that are missing the given output.

        :param predict_id: The id of the prediction.
        """
        return self.missing_outputs(predict_id=predict_id, ids_only=1)

    def _execute_missing_outputs(self, parent):
        """Select the documents that are missing the given output."""
        if len(self.parts[-1][2]) == 0:
            raise ValueError("Predict id is required")
        predict_id = self.parts[-1][2]["predict_id"]
        ids_only = self.parts[-1][2].get('ids_only', False)

        key = f'{CFG.output_prefix}{predict_id}'

        lookup = [
            {
                '$lookup': {
                    'from': key,
                    'localField': '_id',
                    'foreignField': '_source',
                    'as': key,
                }
            },
            {'$match': {key: {'$size': 0}}},
        ]

        raw_cursor = getattr(parent, 'aggregate')(lookup)

        def get_ids(result):
            return {"_id": result["_id"]}

        return SuperDuperCursor(
            raw_cursor=raw_cursor,
            db=self.db,
            id_field='_id',
            process_func=get_ids if ids_only else self._postprocess_result,
            schema=self._get_schema(),
        )

    @property
    @applies_to('find')
    def select_single_id(self, id: str):
        """Return a query that selects a single id.

        :param id: The id to select.
        """
        args, kwargs = self.parts[0][1:]
        args = list(self.args)[:]
        if not args:
            args[0] = {}
        args[0]['_id'] = ObjectId(id)
        return type(self)(
            db=self.db, table=self.table, parts=[('find_one', args, kwargs)]
        )

    @property
    def select_table(self):
        """Return the table or collection to select from."""
        return self.table_or_collection.find()

    def model_update(
        self,
        ids: t.List[t.Any],
        predict_id: str,
        outputs: t.Sequence[t.Any],
        flatten: bool = False,
        **kwargs,
    ):
        """Update the model outputs in the database.

        :param ids: The ids of the documents to update.
        :param predict_id: The id of the prediction.
        :param outputs: The outputs to store.
        :param flatten: Whether to flatten the outputs.
        :param kwargs: Additional keyword arguments.
        """
        if flatten:
            flattened_outputs = []
            flattened_ids = []
            for output, id in zip(outputs, ids):
                assert isinstance(output, (list, tuple)), 'Expected list or tuple'
                for o in output:
                    flattened_outputs.append(o)
                    flattened_ids.append(id)
            return self.model_update(
                ids=flattened_ids,
                predict_id=predict_id,
                outputs=flattened_outputs,
                flatten=False,
                **kwargs,
            )

        documents = []
        for output, id in zip(outputs, ids):
            documents.append(
                {
                    f'{CFG.output_prefix}{predict_id}': output,
                    '_source': ObjectId(id),
                }
            )

        from superduper.base.datalayer import Datalayer

        assert isinstance(self.db, Datalayer)
        output_query = self.db[f'{CFG.output_prefix}{predict_id}'].insert_many(
            documents
        )
        output_query.is_output_query = True
        output_query.updated_key = predict_id
        return output_query

    def _replace_part(self, part_name, replace_function):
        parts = copy.deepcopy(self.parts)

        for i, part in enumerate(parts):
            if part[0] == part_name:
                parts[i] = replace_function(part)
                break

        return type(self)(
            db=self.db,
            table=self.table,
            parts=parts,
        )

    def _execute_select(self, parent):
        parts = self._get_select_parts()
        output = self._get_chain_native_query(parent, parts, method='unpack')
        import mongomock
        import pymongo

        if isinstance(output, (pymongo.cursor.Cursor, mongomock.collection.Cursor)):
            return SuperDuperCursor(
                raw_cursor=output,
                db=self.db,
                id_field='_id',
                schema=self._get_schema(),
            )
        return output

    def _get_select_parts(self):
        def process_select_part(part):
            # Convert the select part to a find part
            _, args, _ = part
            projection = {key: 1 for key in args}
            if projection and '_id' not in projection:
                projection['_id'] = 0
            return ('find', ({}, projection), {})

        def process_find_part(part):
            method, args, kwargs = part
            # args: (filter, projection, *args)
            filter = copy.deepcopy(args[0]) if len(args) > 0 else {}
            filter = dict(filter)
            filter.update(self._get_filter_conditions())
            args = tuple((filter, *args[1:]))

            return (method, args, kwargs)

        parts = []
        for part in self.parts:
            if part[0] == 'select':
                part = process_select_part(part)

            if part[0] in {'find', 'find_one'}:
                part = process_find_part(part)

            if part[0] == 'filter':
                continue
            parts.append(part)

        return parts

    def _get_filter_conditions(self):
        filters = {}
        for part in self.parts:
            if part[0] == 'filter':
                sub_filters = part[1][0]
                assert isinstance(sub_filters, dict)
                filters.update(sub_filters)
        return filters

    def isin(self, other):
        """Create an isin query.

        :param other: The value to check against.
        """
        other = [ObjectId(o) for o in other]
        return self._ops('isin', other)

    def _get_method_parameters(self, method):
        args, kwargs = (), {}
        for part in self.parts:
            if part[0] == method:
                assert not args, 'Multiple find operations found'
                assert not kwargs, 'Multiple find operations found'
                args, kwargs = part[1:]

        return args, kwargs

    def _get_predict_ids(self):
        outputs_parts = [p for p in self.parts if p[0] == 'outputs']
        predict_ids = sum([p[1] for p in outputs_parts], ())
        return predict_ids

    def _execute_outputs(self, parent, method='encode'):
        project = self._get_project()

        limit_args, _ = self._get_method_parameters('limit')
        limit = {"$limit": limit_args[0]} if limit_args else None

        pipeline = []
        filter_mapping_base, filter_mapping_outputs = self._get_filter_mapping()
        if filter_mapping_base:
            pipeline.append({"$match": filter_mapping_base})
            project.update({k: 1 for k in filter_mapping_base.keys()})

        predict_ids_in_filter = list(filter_mapping_outputs.keys())

        predict_ids = self._get_predict_ids()
        predict_ids = list(set(predict_ids).union(predict_ids_in_filter))
        # After the join, the complete outputs data can be queried as
        # {CFG.output_prefix}{predict_id}._outputs.{predict_id} : result.
        for predict_id in predict_ids:
            key = f'{CFG.output_prefix}{predict_id}'
            lookup = {
                "$lookup": {
                    "from": key,
                    "localField": "_id",
                    "foreignField": "_source",
                    "as": key,
                }
            }

            project[key] = 1
            pipeline.append(lookup)

            if predict_id in filter_mapping_outputs:
                filter_key, filter_value = list(
                    filter_mapping_outputs[predict_id].items()
                )[0]
                pipeline.append({"$match": {f'{key}.{filter_key}': filter_value}})

            pipeline.append(
                {"$unwind": {"path": f"${key}", "preserveNullAndEmptyArrays": True}}
            )

        if project:
            pipeline.append({"$project": project})

        if limit:
            pipeline.append(limit)

        try:
            import json

            logging.debug(f'Executing pipeline: {json.dumps(pipeline, indent=2)}')
        except TypeError:
            pass

        raw_cursor = getattr(parent, 'aggregate')(pipeline)

        return SuperDuperCursor(
            raw_cursor=raw_cursor,
            db=self.db,
            id_field='_id',
            process_func=self._postprocess_result,
            schema=self._get_schema(),
        )

    def _get_schema(self):
        outputs_parts = [p for p in self.parts if p[0] == 'outputs']
        predict_ids = sum([p[1] for p in outputs_parts], ())

        try:
            table = self.db.load('table', self.table)
            if not predict_ids:
                return table.schema
            fields = table.schema.fields
        except FileNotFoundError:
            fields = {}

        for predict_id in predict_ids:
            key = f'{CFG.output_prefix}{predict_id}'
            try:
                output_table = self.db.load('table', key)
            except FileNotFoundError:
                logging.warn(
                    f'No schema found for table {key}. Using default projection'
                )
                continue
            fields[key] = output_table.schema.fields[key]

        from superduper.components.datatype import BaseDataType
        from superduper.components.schema import Schema

        fields = {k: v for k, v in fields.items() if isinstance(v, BaseDataType)}

        return Schema(f"_tmp:{self.table}", fields=fields, db=self.db)

    def _get_project(self):
        find_params, _ = self._get_method_parameters('find')
        select_params, _ = self._get_method_parameters('select')
        if self._is_select_find:
            project = {key: 1 for key in select_params}
        else:
            project = copy.deepcopy(find_params[1]) if len(find_params) > 1 else {}

        if not project:
            try:
                table = self.db.load('table', self.table)
                project = {key: 1 for key in table.schema.fields.keys()}
            except FileNotFoundError:
                logging.warn(
                    'No schema found for table',
                    f'{self.table}. Using default projection',
                )

        return project

    def _get_filter_mapping(self):
        find_params, _ = self._get_method_parameters('find')
        filter = find_params[0] if find_params else {}
        filter.update(self._get_filter_conditions())

        if not filter:
            return {}, {}

        filter_mapping_base = {}
        filter_mapping_outputs = defaultdict(dict)

        for key, value in filter.items():
            if '{CFG.output_prefix}' not in key:
                filter_mapping_base[key] = value
                continue

            if key.startswith('{CFG.output_prefix}'):
                predict_id = key.split('__')[1]
                filter_mapping_outputs[predict_id] = {key: value}

        return filter_mapping_base, filter_mapping_outputs

    def _postprocess_result(self, result):
        """Postprocess the result of the query.

        Merge the outputs from the output keys to the result

        :param result: The result to postprocess.
        """
        merge_outputs = {}
        predict_ids = self._get_predict_ids()
        output_keys = [f"{CFG.output_prefix}{predict_id}" for predict_id in predict_ids]
        for output_key in output_keys:
            output_data = result[output_key]
            output_result = output_data[output_key]
            merge_outputs[output_key] = output_result

        result.update(merge_outputs)
        return result


def InsertOne(**kwargs):
    """InsertOne operation for MongoDB.

    :param kwargs: The arguments to pass to the operation.
    """
    return BulkOp(identifier='InsertOne', kwargs=kwargs)


def UpdateOne(**kwargs):
    """UpdateOne operation for MongoDB.

    :param kwargs: The arguments to pass to the operation.
    """
    try:
        filter = kwargs['filter']
    except Exception as e:
        raise KeyError('Filter not found in `UpdateOne`') from e

    id = filter['_id']
    if isinstance(id, ObjectId):
        ids = [id]
    else:
        ids = id['$in']
    kwargs['arg_ids'] = ids
    return BulkOp(identifier='UpdateOne', kwargs=kwargs)


def DeleteOne(**kwargs):
    """DeleteOne operation for MongoDB.

    :param kwargs: The arguments to pass to the operation.
    """
    return BulkOp(identifier='DeleteOne', kwargs=kwargs)


def ReplaceOne(**kwargs):
    """ReplaceOne operation for MongoDB.

    :param kwargs: The arguments to pass to the operation.
    """
    return BulkOp(identifier='ReplaceOne', kwargs=kwargs)


class BulkOp(Leaf):
    """A bulk operation for MongoDB.

    :param kwargs: The arguments to pass to the operation.
    """

    ops: t.ClassVar[t.Sequence[str]] = [
        'InsertOne',
        'UpdateOne',
        'DeleteOne',
        'ReplaceOne',
    ]
    kwargs: t.Dict = dc.field(default_factory=dict)

    def __post_init__(self, db):
        super().__post_init__(db)
        assert self.identifier in self.ops

    @property
    def op(self):
        """Return the operation."""
        kwargs = copy.deepcopy(self.kwargs)
        kwargs.pop('arg_ids')
        for k, v in kwargs.items():
            if isinstance(v, Document):
                kwargs[k] = v.unpack()
        return getattr(pymongo, self.identifier)(**kwargs)



import typing as t

import click
from pymongo.results import DeleteResult, InsertOneResult, UpdateResult
from superduper import logging
from superduper.backends.base.metadata import MetaDataStore
from superduper.components.component import Status
from superduper.misc.colors import Colors


class MongoMetaDataStore(MetaDataStore):
    """
    Metadata store for MongoDB.

    :param uri: URI to the databackend database.
    :param flavour: Flavour of the databackend.
    :param callback: Optional callback to create connection.
    """

    def __init__(
        self,
        uri: t.Optional[str] = None,
        flavour: t.Optional[str] = None,
        callback: t.Optional[t.Callable] = None,
    ):
        super().__init__(uri=uri, flavour=flavour)

        if callback:
            self.connection_callback = callback
        else:
            assert uri
            from .utils import connection_callback

            self.connection_callback = lambda: connection_callback(uri, flavour)

        self.conn, self.name = self.connection_callback()
        self._setup()

    def _setup(self):
        self.db = self.conn[self.name]
        self.meta_collection = self.db['_meta']
        self.component_collection = self.db['_objects']
        self.job_collection = self.db['_jobs']
        self.parent_child_mappings = self.db['_parent_child_mappings']
        self.artifact_relations = self.db['_artifact_relations']

    def reconnect(self):
        """Reconnect to metdata store."""
        self.conn, self.name = self.connection_callback()
        self._setup()

    def url(self):
        """Metadata store connection url."""
        return self.conn.HOST + ':' + str(self.conn.PORT) + '/' + self.name

    def drop(self, force: bool = False):
        """Drop all meta-data from the metadata store.

        Please always use with caution. This will drop all the meta-data collections.
        :param force: whether to force the drop, defaults to False
        """
        if not force:
            if not click.confirm(
                f'{Colors.RED}[!!!WARNING USE WITH CAUTION AS YOU '
                f'WILL LOSE ALL DATA!!!]{Colors.RESET} '
                'Are you sure you want to drop all meta-data? ',
                default=False,
            ):
                logging.warn('Aborting...')
        self.db.drop_collection(self.meta_collection.name)
        self.db.drop_collection(self.component_collection.name)
        self.db.drop_collection(self.job_collection.name)
        self.db.drop_collection(self.parent_child_mappings.name)
        self.db.drop_collection(self.artifact_relations.name)

    def delete_parent_child(self, parent: str, child: str) -> None:
        """
        Delete parent-child relationships between two components.

        :param parent: parent component uuid
        :param child: child component uuid
        """
        self.parent_child_mappings.delete_many(
            {
                'parent_id': parent,
                'child_id': child,
            }
        )

    def create_parent_child(self, parent: str, child: str) -> None:
        """Create a parent-child relationship between two components.

        :param parent: parent component
        :param child: child component
        """
        self.parent_child_mappings.insert_one(
            {
                'parent_id': parent,
                'child_id': child,
            }
        )

    def _create_data(self, table_name, datas):
        collection = self.db[table_name]
        collection.insert_many(datas)

    def _delete_data(self, table_name, filter):
        collection = self.db[table_name]
        collection.delete_many(filter)

    def _get_data(self, table_name, filter):
        collection = self.db[table_name]
        return list(collection.find(filter))

    def create_component(self, info: t.Dict) -> InsertOneResult:
        """Create a component in the metadata store.

        :param info: dictionary containing information about the component.
        """
        if 'hidden' not in info:
            info['hidden'] = False
        return self.component_collection.insert_one(info)

    def create_job(self, info: t.Dict) -> InsertOneResult:
        """Create a job in the metadata store.

        :param info: dictionary containing information about the job.
        """
        return self.job_collection.insert_one(info)

    def get_job(self, identifier: str):
        """Get a job from the metadata store.

        :param identifier: identifier of job
        """
        return self.job_collection.find_one({'job_id': identifier})

    def get_latest_version(
        self, type_id: str, identifier: str, allow_hidden: bool = False
    ) -> int:
        """Get the latest version of a component.

        :param type_id: type of component
        :param identifier: identifier of component
        :param allow_hidden: whether to allow hidden components
        """
        try:
            if allow_hidden:
                return sorted(
                    self.component_collection.distinct(
                        'version', {'identifier': identifier, 'type_id': type_id}
                    )
                )[-1]
            else:

                def f():
                    return self.component_collection.distinct(
                        'version',
                        {
                            '$or': [
                                {
                                    'identifier': identifier,
                                    'type_id': type_id,
                                    'hidden': False,
                                },
                                {
                                    'identifier': identifier,
                                    'type_id': type_id,
                                    'hidden': {'$exists': 0},
                                },
                            ]
                        },
                    )

                return sorted(f())[-1]
        except IndexError:
            raise FileNotFoundError(f'Can\'t find {type_id}: {identifier} in metadata')

    def update_job(self, identifier: str, key: str, value: t.Any) -> UpdateResult:
        """Update a job in the metadata store.

        :param identifier: identifier of job
        :param key: key to be updated
        :param value: value to be updated
        """
        return self.job_collection.update_one(
            {'job_id': identifier}, {'$set': {key: value}}
        )

    def show_cdc_tables(self):
        """Show tables to be consumed with cdc."""
        return self.component_collection.distinct('cdc_table')

    def _show_cdcs(self, table):
        return list(
            self.component_collection.find(
                {'cdc_table': table},
                {'identifier': 1, '_id': 0, 'type_id': 1, 'version': 1, 'uuid': 1},
            )
        )

    def _show_components(self, type_id: t.Optional[str] = None):
        """Show components in the metadata store.

        :param type_id: type of component
        """
        filter = {}
        if type_id is not None:
            filter['type_id'] = type_id
        return list(
            self.component_collection.find(
                filter, {'identifier': 1, '_id': 0, 'type_id': 1}
            )
        )

    def show_component_versions(
        self, type_id: str, identifier: str
    ) -> t.List[t.Union[t.Any, int]]:
        """Show component versions in the metadata store.

        :param type_id: type of component
        :param identifier: identifier of component
        """
        return sorted(
            self.component_collection.distinct(
                'version', {'type_id': type_id, 'identifier': identifier}
            )
        )

    def show_job_ids(self, uuids: t.Optional[str] = None, status: str = 'running'):
        """Show all jobs in the metadata store.

        :param uuids: list of uuids
        :param status: status of the job
        """
        if uuids:
            return self.job_collection.distinct(
                'job_id', {'status': status, 'uuid': {'$in': uuids}}
            )
        return self.job_collection.distinct('job_id', {'status': status})

    def show_jobs(
        self,
        identifier: t.Optional[str] = None,
        type_id: t.Optional[str] = None,
    ):
        """Show jobs in the metadata store.

        :param component_identifier: identifier of component
        :param type_id: type of component
        """
        filter_ = {}
        if identifier is not None:
            filter_['identifier'] = identifier
        if type_id is not None:
            filter_['type_id'] = type_id
        return list(self.job_collection.find(filter_))

    def _get_component_uuid(self, type_id: str, identifier: str, version: int) -> str:
        return self.component_collection.find_one(
            {'type_id': type_id, 'identifier': identifier, 'version': version},
            {'uuid': 1},
        )['uuid']

    def component_version_has_parents(
        self, type_id: str, identifier: str, version: int
    ) -> int:
        """Check if a component version has parents.

        :param type_id: type of component
        :param identifier: identifier of component
        :param version: version of component
        """
        uuid = self.component_collection.find_one(
            {'type_id': type_id, 'identifier': identifier, 'version': version},
            {'uuid': 1, 'id': 1},
        )['uuid']
        doc = {'child_id': uuid}
        return self.parent_child_mappings.count_documents(doc)

    def delete_component_version(
        self, type_id: str, identifier: str, version: int
    ) -> DeleteResult:
        """Delete a component version from the metadata store.

        :param type_id: type of component
        :param identifier: identifier of component
        :param version: version of component
        """
        uuid = self._get_component_uuid(type_id, identifier, version)

        self.parent_child_mappings.delete_many({'parent': uuid})

        return self.component_collection.delete_many(
            {
                'identifier': identifier,
                'type_id': type_id,
                'version': version,
            }
        )

    def get_component_by_uuid(self, uuid: str, allow_hidden: bool = False):
        """Get a component by UUID.

        :param uuid: UUID of component
        :param allow_hidden: whether to load hidden components
        """
        r = self.component_collection.find_one({'uuid': uuid})
        if r is None:
            raise FileNotFoundError(f'Can\'t find {uuid} in metadata')
        return self._get_component(
            type_id=r['type_id'],
            identifier=r['identifier'],
            version=r['version'],
            allow_hidden=allow_hidden,
        )

    def _get_component(
        self,
        type_id: str,
        identifier: str,
        version: int,
        allow_hidden: bool = False,
    ) -> t.Dict[str, t.Any]:
        if not allow_hidden:
            r = self.component_collection.find_one(
                {
                    '$or': [
                        {
                            'identifier': identifier,
                            'type_id': type_id,
                            'version': version,
                            'hidden': False,
                        },
                        {
                            'identifier': identifier,
                            'type_id': type_id,
                            'version': version,
                            'hidden': {'$exists': 0},
                        },
                    ]
                },
                {'_id': 0},
            )
        else:
            r = self.component_collection.find_one(
                {
                    'identifier': identifier,
                    'type_id': type_id,
                    'version': version,
                },
                {'_id': 0},
            )
        return r

    def get_component_version_parents(self, uuid: str) -> t.List[str]:
        """Get the parents of a component version.

        :param uuid: unique identifier of component
        """
        return [
            r['parent_id'] for r in self.parent_child_mappings.find({'child_id': uuid})
        ]

    def _replace_object(
        self,
        info: t.Dict[str, t.Any],
        identifier: str | None = None,
        type_id: str | None = None,
        version: int | None = None,
        uuid: str | None = None,
    ) -> None:
        if uuid:
            return self.component_collection.replace_one({'uuid': uuid}, info)
        else:
            assert type_id, 'type_id cannot be False in replace'
            assert identifier, "identifier cannot be False in replace"
            assert version is not None, 'version cannot be None in replace'
            self.component_collection.replace_one(
                {'identifier': identifier, 'type_id': type_id, 'version': version},
                info,
            )

    def _update_object(
        self,
        identifier: str,
        type_id: str,
        key: str,
        value: t.Any,
        version: int,
    ):
        """Update a component in the metadata store.

        :param identifier: identifier of component
        :param type_id: type of component
        :param key: key to be updated
        :param value: value to be updated
        :param version: version of component
        """
        return self.component_collection.update_one(
            {'identifier': identifier, 'type_id': type_id, 'version': version},
            {'$set': {key: value}},
        )

    def hide_component_version(
        self, type_id: str, identifier: str, version: int
    ) -> None:
        """Hide a component version.

        :param type_id: type of component
        :param identifier: identifier of component
        :param version: version of component
        """
        self.component_collection.update_one(
            {'type_id': type_id, 'identifier': identifier, 'version': version},
            {'$set': {'hidden': True}},
        )

    def disconnect(self):
        """Disconnect the client."""

        # TODO: implement me

    def set_component_status(self, uuid, status: Status):
        """Set status of component with `status`."""
        return self.component_collection.update_one(
            {'uuid': uuid}, {'$set': {'status': status}}
        )

    def _get_component_status(self, uuid):
        """Get status of component."""
        data = self.component_collection.find_one({'uuid': uuid}, {'status': 1})
        return data['status'] if data else None
